@@ -32,6 +32,7 @@
 
 #include "../../events/imKStoUCS.h"
 #include "../../events/SDL_keysym_to_scancode_c.h"
+#include "../../events/SDL_keysym_to_keycode_c.h"
 
 #ifdef X_HAVE_UTF8_STRING
 #include <locale.h>
@@ -90,9 +91,9 @@ KeySym X11_KeyCodeToSym(SDL_VideoDevice *_this, KeyCode keycode, unsigned char g
     SDL_zero(mods_ret);
 
 #ifdef SDL_VIDEO_DRIVER_X11_HAS_XKBLOOKUPKEYSYM
-    if (data->xkb) {
-        int num_groups = XkbKeyNumGroups(data->xkb, keycode);
-        unsigned char info = XkbKeyGroupInfo(data->xkb, keycode);
+    if (data->xkb.desc_ptr) {
+        int num_groups = XkbKeyNumGroups(data->xkb.desc_ptr, keycode);
+        unsigned char info = XkbKeyGroupInfo(data->xkb.desc_ptr, keycode);
 
         if (num_groups && group >= num_groups) {
 
@@ -152,8 +153,8 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
         int xkb_major = XkbMajorVersion;
         int xkb_minor = XkbMinorVersion;
 
-        if (X11_XkbQueryExtension(data->display, NULL, &data->xkb_event, NULL, &xkb_major, &xkb_minor)) {
-            data->xkb = X11_XkbGetMap(data->display, XkbAllClientInfoMask, XkbUseCoreKbd);
+        if (X11_XkbQueryExtension(data->display, NULL, &data->xkb.event, NULL, &xkb_major, &xkb_minor)) {
+            data->xkb.desc_ptr = X11_XkbGetMap(data->display, XkbAllClientInfoMask, XkbUseCoreKbd);
         }
 
         // This will remove KeyRelease events for held keys
@@ -168,9 +169,6 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
            Compose keys will work correctly. */
         char *prev_locale = setlocale(LC_ALL, NULL);
         char *prev_xmods = X11_XSetLocaleModifiers(NULL);
-        const char *new_xmods = "";
-        const char *env_xmods = SDL_getenv("XMODIFIERS");
-        bool has_dbus_ime_support = false;
 
         if (prev_locale) {
             prev_locale = SDL_strdup(prev_locale);
@@ -180,22 +178,8 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
             prev_xmods = SDL_strdup(prev_xmods);
         }
 
-        /* IBus resends some key events that were filtered by XFilterEvents
-           when it is used via XIM which causes issues. Prevent this by forcing
-           @im=none if XMODIFIERS contains @im=ibus. IBus can still be used via
-           the DBus implementation, which also has support for pre-editing. */
-        if (env_xmods && SDL_strstr(env_xmods, "@im=ibus") != NULL) {
-            has_dbus_ime_support = true;
-        }
-        if (env_xmods && SDL_strstr(env_xmods, "@im=fcitx") != NULL) {
-            has_dbus_ime_support = true;
-        }
-        if (has_dbus_ime_support || !xkb_repeat) {
-            new_xmods = "@im=none";
-        }
-
         (void)setlocale(LC_ALL, "");
-        X11_XSetLocaleModifiers(new_xmods);
+        X11_XSetLocaleModifiers("");
 
         data->im = X11_XOpenIM(data->display, NULL, NULL, NULL);
 
@@ -255,7 +239,7 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
         const SDL_Scancode *table = SDL_GetScancodeTable(scancode_set[best_index], &table_size);
 
 #ifdef DEBUG_KEYBOARD
-        SDL_Log("Using scancode set %d, min_keycode = %d, max_keycode = %d, table_size = %d\n", best_index, min_keycode, max_keycode, table_size);
+        SDL_Log("Using scancode set %d, min_keycode = %d, max_keycode = %d, table_size = %d", best_index, min_keycode, max_keycode, table_size);
 #endif
         // This should never happen, but just in case...
         if (table_size > (SDL_arraysize(data->key_layout) - min_keycode)) {
@@ -280,17 +264,17 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
             if (scancode == data->key_layout[i]) {
                 continue;
             }
-            if ((SDL_GetKeymapKeycode(NULL, scancode, SDL_KMOD_NONE) & SDLK_SCANCODE_MASK) && X11_ScancodeIsRemappable(scancode)) {
+            if ((SDL_GetKeymapKeycode(NULL, scancode, SDL_KMOD_NONE) & (SDLK_SCANCODE_MASK | SDLK_EXTENDED_MASK)) && X11_ScancodeIsRemappable(scancode)) {
                 // Not a character key and the scancode is safe to remap
 #ifdef DEBUG_KEYBOARD
-                SDL_Log("Changing scancode, was %d (%s), now %d (%s)\n", data->key_layout[i], SDL_GetScancodeName(data->key_layout[i]), scancode, SDL_GetScancodeName(scancode));
+                SDL_Log("Changing scancode, was %d (%s), now %d (%s)", data->key_layout[i], SDL_GetScancodeName(data->key_layout[i]), scancode, SDL_GetScancodeName(scancode));
 #endif
                 data->key_layout[i] = scancode;
             }
         }
     } else {
 #ifdef DEBUG_SCANCODES
-        SDL_Log("Keyboard layout unknown, please report the following to the SDL forums/mailing list (https://discourse.libsdl.org/):\n");
+        SDL_Log("Keyboard layout unknown, please report the following to the SDL forums/mailing list (https://discourse.libsdl.org/):");
 #endif
 
         // Determine key_layout - only works on US QWERTY layout
@@ -304,9 +288,9 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
                         (unsigned int)sym, sym == NoSymbol ? "NoSymbol" : X11_XKeysymToString(sym));
             }
             if (scancode == SDL_SCANCODE_UNKNOWN) {
-                SDL_Log("scancode not found\n");
+                SDL_Log("scancode not found");
             } else {
-                SDL_Log("scancode = %d (%s)\n", scancode, SDL_GetScancodeName(scancode));
+                SDL_Log("scancode = %d (%s)", scancode, SDL_GetScancodeName(scancode));
             }
 #endif
             data->key_layout[i] = scancode;
@@ -317,15 +301,61 @@ bool X11_InitKeyboard(SDL_VideoDevice *_this)
 
     SDL_SetScancodeName(SDL_SCANCODE_APPLICATION, "Menu");
 
-#ifdef SDL_USE_IME
-    SDL_IME_Init();
-#endif
-
     X11_ReconcileKeyboardState(_this);
 
     return true;
 }
 
+static unsigned X11_GetNumLockModifierMask(SDL_VideoDevice *_this)
+{
+    SDL_VideoData *videodata = _this->internal;
+    Display *display = videodata->display;
+    unsigned num_mask = 0;
+    int i, j;
+    XModifierKeymap *xmods;
+    unsigned n;
+
+    xmods = X11_XGetModifierMapping(display);
+    n = xmods->max_keypermod;
+    for (i = 3; i < 8; i++) {
+        for (j = 0; j < n; j++) {
+            KeyCode kc = xmods->modifiermap[i * n + j];
+            if (videodata->key_layout[kc] == SDL_SCANCODE_NUMLOCKCLEAR) {
+                num_mask = 1 << i;
+                break;
+            }
+        }
+    }
+    X11_XFreeModifiermap(xmods);
+
+    return num_mask;
+}
+
+static unsigned X11_GetScrollLockModifierMask(SDL_VideoDevice *_this)
+{
+    SDL_VideoData *videodata = _this->internal;
+    Display *display = videodata->display;
+    unsigned num_mask = 0;
+    int i, j;
+    XModifierKeymap *xmods;
+    unsigned n;
+
+    xmods = X11_XGetModifierMapping(display);
+    n = xmods->max_keypermod;
+    for (i = 3; i < 8; i++) {
+        for (j = 0; j < n; j++) {
+            KeyCode kc = xmods->modifiermap[i * n + j];
+            if (videodata->key_layout[kc] == SDL_SCANCODE_SCROLLLOCK) {
+                num_mask = 1 << i;
+                break;
+            }
+        }
+    }
+    X11_XFreeModifiermap(xmods);
+
+    return num_mask;
+}
+
 void X11_UpdateKeymap(SDL_VideoDevice *_this, bool send_event)
 {
     struct Keymod_masks
@@ -340,53 +370,72 @@ void X11_UpdateKeymap(SDL_VideoDevice *_this, bool send_event)
         { SDL_KMOD_MODE, Mod5Mask },
         { SDL_KMOD_MODE | SDL_KMOD_SHIFT, Mod5Mask | ShiftMask },
         { SDL_KMOD_MODE | SDL_KMOD_CAPS, Mod5Mask | LockMask },
-        { SDL_KMOD_MODE | SDL_KMOD_SHIFT | SDL_KMOD_CAPS, Mod5Mask | ShiftMask | LockMask }
+        { SDL_KMOD_MODE | SDL_KMOD_SHIFT | SDL_KMOD_CAPS, Mod5Mask | ShiftMask | LockMask },
+        { SDL_KMOD_LEVEL5, Mod3Mask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_SHIFT, Mod3Mask | ShiftMask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_CAPS, Mod3Mask | LockMask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_SHIFT | SDL_KMOD_CAPS, Mod3Mask | ShiftMask | LockMask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_MODE, Mod5Mask | Mod3Mask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_MODE | SDL_KMOD_SHIFT, Mod3Mask | Mod5Mask | ShiftMask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_MODE | SDL_KMOD_CAPS, Mod3Mask | Mod5Mask | LockMask },
+        { SDL_KMOD_LEVEL5 | SDL_KMOD_MODE | SDL_KMOD_SHIFT | SDL_KMOD_CAPS, Mod3Mask | Mod5Mask | ShiftMask | LockMask }
     };
 
     SDL_VideoData *data = _this->internal;
-    int i;
     SDL_Scancode scancode;
-    SDL_Keymap *keymap;
-
-    keymap = SDL_CreateKeymap();
+    SDL_Keymap *keymap = SDL_CreateKeymap();
 
 #ifdef SDL_VIDEO_DRIVER_X11_HAS_XKBLOOKUPKEYSYM
-    if (data->xkb) {
+    if (data->xkb.desc_ptr) {
         XkbStateRec state;
-        X11_XkbGetUpdatedMap(data->display, XkbAllClientInfoMask, data->xkb);
+        X11_XkbGetUpdatedMap(data->display, XkbAllClientInfoMask, data->xkb.desc_ptr);
 
         if (X11_XkbGetState(data->display, XkbUseCoreKbd, &state) == Success) {
-            data->xkb_group = state.group;
+            data->xkb.current_group = state.group;
         }
     }
 #endif
 
     for (int m = 0; m < SDL_arraysize(keymod_masks); ++m) {
-        for (i = 0; i < SDL_arraysize(data->key_layout); i++) {
-            SDL_Keycode keycode;
-
+        for (int i = 0; i < SDL_arraysize(data->key_layout); ++i) {
             // Make sure this is a valid scancode
             scancode = data->key_layout[i];
             if (scancode == SDL_SCANCODE_UNKNOWN) {
                 continue;
             }
 
-            KeySym keysym = X11_KeyCodeToSym(_this, i, data->xkb_group, keymod_masks[m].xkb_mask);
+            const KeySym keysym = X11_KeyCodeToSym(_this, i, data->xkb.current_group, keymod_masks[m].xkb_mask);
 
-            // Note: The default SDL scancode table sets this to right alt instead of AltGr/Mode, so handle it separately.
-            if (keysym != XK_ISO_Level3_Shift) {
-                keycode = SDL_KeySymToUcs4(keysym);
-            } else {
-                keycode = SDLK_MODE;
-            }
+            if (keysym != NoSymbol) {
+                SDL_Keycode keycode = SDL_GetKeyCodeFromKeySym(keysym, i, keymod_masks[m].sdl_mask);
 
-            if (!keycode) {
-                SDL_Scancode keyScancode = SDL_GetScancodeFromKeySym(keysym, (KeyCode)i);
-                keycode = SDL_GetKeymapKeycode(NULL, keyScancode, keymod_masks[m].sdl_mask);
+                if (!keycode) {
+                    switch (scancode) {
+                    case SDL_SCANCODE_RETURN:
+                        keycode = SDLK_RETURN;
+                        break;
+                    case SDL_SCANCODE_ESCAPE:
+                        keycode = SDLK_ESCAPE;
+                        break;
+                    case SDL_SCANCODE_BACKSPACE:
+                        keycode = SDLK_BACKSPACE;
+                        break;
+                    case SDL_SCANCODE_DELETE:
+                        keycode = SDLK_DELETE;
+                        break;
+                    default:
+                        keycode = SDL_SCANCODE_TO_KEYCODE(scancode);
+                        break;
+                    }
+                }
+
+                SDL_SetKeymapEntry(keymap, scancode, keymod_masks[m].sdl_mask, keycode);
             }
-            SDL_SetKeymapEntry(keymap, scancode, keymod_masks[m].sdl_mask, keycode);
         }
     }
+
+    data->xkb.numlock_mask = X11_GetNumLockModifierMask(_this);
+    data->xkb.scrolllock_mask = X11_GetScrollLockModifierMask(_this);
     SDL_SetKeymap(keymap, send_event);
 }
 
@@ -395,15 +444,232 @@ void X11_QuitKeyboard(SDL_VideoDevice *_this)
     SDL_VideoData *data = _this->internal;
 
 #ifdef SDL_VIDEO_DRIVER_X11_HAS_XKBLOOKUPKEYSYM
-    if (data->xkb) {
-        X11_XkbFreeKeyboard(data->xkb, 0, True);
-        data->xkb = NULL;
+    if (data->xkb.desc_ptr) {
+        X11_XkbFreeKeyboard(data->xkb.desc_ptr, 0, True);
+        data->xkb.desc_ptr = NULL;
     }
 #endif
+}
 
-#ifdef SDL_USE_IME
-    SDL_IME_Quit();
+void X11_ClearComposition(SDL_WindowData *data)
+{
+    if (data->preedit_length > 0) {
+        data->preedit_text[0] = '\0';
+        data->preedit_length = 0;
+    }
+
+    if (data->ime_needs_clear_composition) {
+        SDL_SendEditingText("", 0, 0);
+        data->ime_needs_clear_composition = false;
+    }
+}
+
+static void X11_SendEditingEvent(SDL_WindowData *data)
+{
+    if (data->preedit_length == 0) {
+        X11_ClearComposition(data);
+        return;
+    }
+
+    bool in_highlight = false;
+    int start = -1, length = 0, i;
+    for (i = 0; i < data->preedit_length; ++i) {
+        if (data->preedit_feedback[i] & (XIMReverse | XIMHighlight)) {
+            if (start < 0) {
+                start = i;
+                in_highlight = true;
+            }
+        } else if (in_highlight) {
+            // Found the end of the highlight
+            break;
+        }
+    }
+    if (in_highlight) {
+        length = (i - start);
+    } else {
+        start = SDL_clamp(data->preedit_cursor, 0, data->preedit_length);
+    }
+    SDL_SendEditingText(data->preedit_text, start, length);
+
+    data->ime_needs_clear_composition = true;
+}
+
+static int preedit_start_callback(XIC xic, XPointer client_data, XPointer call_data)
+{
+    // No limit on preedit text length
+    return -1;
+}
+
+static void preedit_done_callback(XIC xic, XPointer client_data, XPointer call_data)
+{
+}
+
+static void preedit_draw_callback(XIC xic, XPointer client_data, XIMPreeditDrawCallbackStruct *call_data)
+{
+    SDL_WindowData *data = (SDL_WindowData *)client_data;
+    int chg_first = SDL_clamp(call_data->chg_first, 0, data->preedit_length);
+    int chg_length = SDL_clamp(call_data->chg_length, 0, data->preedit_length - chg_first);
+
+    const char *start = data->preedit_text;
+    if (chg_length > 0) {
+        // Delete text in range
+        for (int i = 0; start && *start && i < chg_first; ++i) {
+            SDL_StepUTF8(&start, NULL);
+        }
+
+        const char *end = start;
+        for (int i = 0; end && *end && i < chg_length; ++i) {
+            SDL_StepUTF8(&end, NULL);
+        }
+
+        if (end > start) {
+            SDL_memmove((char *)start, end, SDL_strlen(end) + 1);
+            if ((chg_first + chg_length) > data->preedit_length) {
+                SDL_memmove(&data->preedit_feedback[chg_first], &data->preedit_feedback[chg_first + chg_length], (data->preedit_length - chg_first - chg_length) * sizeof(*data->preedit_feedback));
+            }
+        }
+        data->preedit_length -= chg_length;
+    }
+
+    XIMText *text = call_data->text;
+    if (text) {
+        // Insert text in range
+        SDL_assert(!text->encoding_is_wchar);
+
+        // The text length isn't calculated as directed by the spec, recalculate it now
+        if (text->string.multi_byte) {
+            text->length = SDL_utf8strlen(text->string.multi_byte);
+        }
+
+        size_t string_size = SDL_strlen(text->string.multi_byte);
+        size_t size = string_size + 1;
+        if (data->preedit_text) {
+            size += SDL_strlen(data->preedit_text);
+        }
+        char *preedit_text = (char *)SDL_malloc(size * sizeof(*preedit_text));
+        if (preedit_text) {
+            size_t pre_size = (start - data->preedit_text);
+            size_t post_size = start ? SDL_strlen(start) : 0;
+            if (pre_size > 0) {
+                SDL_memcpy(&preedit_text[0], data->preedit_text, pre_size);
+            }
+            SDL_memcpy(&preedit_text[pre_size], text->string.multi_byte, string_size);
+            if (post_size > 0) {
+                SDL_memcpy(&preedit_text[pre_size + string_size], start, post_size);
+            }
+            preedit_text[size - 1] = '\0';
+        }
+
+        size_t feedback_size = data->preedit_length + text->length;
+        XIMFeedback *feedback = (XIMFeedback *)SDL_malloc(feedback_size * sizeof(*feedback));
+        if (feedback) {
+            size_t pre_size = (size_t)chg_first;
+            size_t post_size = (size_t)data->preedit_length - pre_size;
+            if (pre_size > 0) {
+                SDL_memcpy(&feedback[0], data->preedit_feedback, pre_size * sizeof(*feedback));
+            }
+            SDL_memcpy(&feedback[pre_size], text->feedback, text->length * sizeof(*feedback));
+            if (post_size > 0) {
+                SDL_memcpy(&feedback[pre_size + text->length], &data->preedit_feedback[pre_size], post_size * sizeof(*feedback));
+            }
+        }
+
+        if (preedit_text && feedback) {
+            SDL_free(data->preedit_text);
+            data->preedit_text = preedit_text;
+
+            SDL_free(data->preedit_feedback);
+            data->preedit_feedback = feedback;
+
+            data->preedit_length += text->length;
+        } else {
+            SDL_free(preedit_text);
+            SDL_free(feedback);
+        }
+    }
+
+    data->preedit_cursor = call_data->caret;
+
+#ifdef DEBUG_XIM
+    if (call_data->chg_length > 0) {
+        SDL_Log("Draw callback deleted %d characters at %d", call_data->chg_length, call_data->chg_first);
+    }
+    if (text) {
+        SDL_Log("Draw callback inserted %s at %d, caret: %d", text->string.multi_byte, call_data->chg_first, call_data->caret);
+    }
+    SDL_Log("Pre-edit text: %s", data->preedit_text);
 #endif
+
+    X11_SendEditingEvent(data);
+}
+
+static void preedit_caret_callback(XIC xic, XPointer client_data, XIMPreeditCaretCallbackStruct *call_data)
+{
+    SDL_WindowData *data = (SDL_WindowData *)client_data;
+
+    switch (call_data->direction) {
+    case XIMAbsolutePosition:
+        if (call_data->position != data->preedit_cursor) {
+            data->preedit_cursor = call_data->position;
+            X11_SendEditingEvent(data);
+        }
+        break;
+    case XIMDontChange:
+        break;
+    default:
+        // Not currently supported
+        break;
+    }
+}
+
+void X11_CreateInputContext(SDL_WindowData *data)
+{
+#ifdef X_HAVE_UTF8_STRING
+    SDL_VideoData *videodata = data->videodata;
+
+    if (SDL_X11_HAVE_UTF8 && videodata->im) {
+        const char *hint = SDL_GetHint(SDL_HINT_IME_IMPLEMENTED_UI);
+        if (hint && SDL_strstr(hint, "composition")) {
+            XIMCallback draw_callback;
+            draw_callback.client_data = (XPointer)data;
+            draw_callback.callback = (XIMProc)preedit_draw_callback;
+
+            XIMCallback start_callback;
+            start_callback.client_data = (XPointer)data;
+            start_callback.callback = (XIMProc)preedit_start_callback;
+
+            XIMCallback done_callback;
+            done_callback.client_data = (XPointer)data;
+            done_callback.callback = (XIMProc)preedit_done_callback;
+
+            XIMCallback caret_callback;
+            caret_callback.client_data = (XPointer)data;
+            caret_callback.callback = (XIMProc)preedit_caret_callback;
+
+            XVaNestedList attr = X11_XVaCreateNestedList(0,
+                                                         XNPreeditStartCallback, &start_callback,
+                                                         XNPreeditDoneCallback, &done_callback,
+                                                         XNPreeditDrawCallback, &draw_callback,
+                                                         XNPreeditCaretCallback, &caret_callback,
+                                                         NULL);
+            if (attr) {
+                data->ic = X11_XCreateIC(videodata->im,
+                                         XNInputStyle, XIMPreeditCallbacks | XIMStatusCallbacks,
+                                         XNPreeditAttributes, attr,
+                                         XNClientWindow, data->xwindow,
+                                         NULL);
+                X11_XFree(attr);
+            }
+        } else {
+            data->ic = X11_XCreateIC(videodata->im,
+                                     XNInputStyle, XIMPreeditNothing | XIMStatusNothing,
+                                     XNClientWindow, data->xwindow,
+                                     NULL);
+        }
+        data->xim_spot.x = -1;
+        data->xim_spot.y = -1;
+    }
+#endif // X_HAVE_UTF8_STRING
 }
 
 static void X11_ResetXIM(SDL_VideoDevice *_this, SDL_Window *window)
@@ -418,7 +684,7 @@ static void X11_ResetXIM(SDL_VideoDevice *_this, SDL_Window *window)
             X11_XFree(contents);
         }
     }
-#endif
+#endif // X_HAVE_UTF8_STRING
 }
 
 bool X11_StartTextInput(SDL_VideoDevice *_this, SDL_Window *window, SDL_PropertiesID props)
@@ -431,16 +697,27 @@ bool X11_StartTextInput(SDL_VideoDevice *_this, SDL_Window *window, SDL_Properti
 bool X11_StopTextInput(SDL_VideoDevice *_this, SDL_Window *window)
 {
     X11_ResetXIM(_this, window);
-#ifdef SDL_USE_IME
-    SDL_IME_Reset();
-#endif
     return true;
 }
 
 bool X11_UpdateTextInputArea(SDL_VideoDevice *_this, SDL_Window *window)
 {
-#ifdef SDL_USE_IME
-    SDL_IME_UpdateTextInputArea(window);
+#ifdef X_HAVE_UTF8_STRING
+    SDL_WindowData *data = window->internal;
+
+    if (data && data->ic) {
+        XPoint spot;
+        spot.x = window->text_input_rect.x + window->text_input_cursor;
+        spot.y = window->text_input_rect.y + window->text_input_rect.h;
+        if (spot.x != data->xim_spot.x || spot.y != data->xim_spot.y) {
+            XVaNestedList attr = X11_XVaCreateNestedList(0, XNSpotLocation, &spot, NULL);
+            if (attr) {
+                X11_XSetICValues(data->ic, XNPreeditAttributes, attr, NULL);
+                X11_XFree(attr);
+            }
+            SDL_copyp(&data->xim_spot, &spot);
+        }
+    }
 #endif
     return true;
 }
